from typing import Tuple

import numpy as np

from entity_embeddings.config import Config
from entity_embeddings.network.network import EmbeddingNetwork
from entity_embeddings.util import model_utils, preprocessing_utils


class Embedder():
    def __init__(self, config: Config):
        self.config = config
        self.network = EmbeddingNetwork(config)

    def prepare_data(self, config: Config) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        X, y = preprocessing_utils.get_X_y(config.df, config.target_name)

        # pre processing of X and Y
        X = preprocessing_utils.label_encode(X)
        y = np.array(y)

        num_records = len(X)
        train_size = int(config.train_ratio * num_records)

        X_train = X[:train_size]
        X_val = X[train_size:]
        y_train = y[:train_size]
        y_val = y[train_size:]

        X_train, y_train = preprocessing_utils.sample(X_train, y_train, 1000)  # Simulate data sparsity

        y_train = config.target_processor.process_target(y_train.tolist())
        y_val = config.target_processor.process_target(y_val.tolist())

        return X_train, X_val, y_train, y_val

    def perform_embedding(self) -> None:
        self.X_train, self.X_val, self.y_train, self.y_val = self.prepare_data(self.config)
        self.network.fit(self.X_train, self.y_train, self.X_val, self.y_val)

        model_utils.save_weights(self.network.model, self.config)



from abc import ABC, abstractmethod
from typing import List

import numpy as np
import pandas as pd
from sklearn.preprocessing import LabelEncoder, OneHotEncoder, MinMaxScaler


class TargetProcessor(ABC):
    @abstractmethod
    def process_target(self, y: List) -> np.ndarray:
        pass


class BinaryClassificationProcessor(TargetProcessor):
    def process_target(self, y: List) -> np.ndarray:
        return LabelEncoder().fit_transform(y)


class MulticlassClassificationProcessor(TargetProcessor):
    def process_target(self, y: List) -> np.ndarray:
        return OneHotEncoder().fit_transform(pd.DataFrame(y))


class RegressionProcessor(TargetProcessor):
    def process_target(self, y: List) -> np.ndarray:
        return MinMaxScaler().fit_transform(pd.DataFrame(y))



from entity_embeddings.processor import processor, RegressionProcessor, MulticlassClassificationProcessor, \
    BinaryClassificationProcessor
from entity_embeddings.processor.target_type import TargetType


def get_target_processor(type: int) -> processor:
    """
    This method is used to get a target processor from a given target type
    :param type:
    :return:
    """
    if type == TargetType.REGRESSION:
        return RegressionProcessor()
    elif type == TargetType.BINARY_CLASSIFICATION:
        return BinaryClassificationProcessor()
    elif type == TargetType.MULTICLASS_CLASSIFICATION:
        return MulticlassClassificationProcessor()

    raise ValueError("You should provide a valid target type")



import unittest
from typing import List

import numpy as np
import pandas as pd
from keras.engine import Layer
from keras.layers import Activation, Dense, Concatenate
from sklearn.preprocessing import LabelEncoder

from entity_embeddings.config import Config, get_embedding_size
from entity_embeddings.network import ModelAssembler
from entity_embeddings.processor.processor import TargetProcessor
from entity_embeddings.processor.target_type import TargetType
from entity_embeddings.util.dataframe_utils import create_random_csv, remove_random_csv


class TestConfig(unittest.TestCase):
    def test_default_config(self):
        random_csv = create_random_csv()
        target = 'D'

        config = Config.make_default_config(csv_path=random_csv,
                                            target_name=target,
                                            target_type=TargetType.BINARY_CLASSIFICATION,
                                            train_ratio=0.9)
        self.assertIsNotNone(config)
        remove_random_csv()

    def test_custom_config(self):
        random_csv = create_random_csv()
        target = 'D'

        processor = CustomProcessor()
        assembler = CustomAssembler()

        config = Config.make_custom_config(csv_path=random_csv,
                                           target_name=target,
                                           train_ratio=0.9,
                                           target_processor=processor,
                                           model_assembler=assembler)

        self.assertIsNotNone(config)
        remove_random_csv()

    def test_embedding_size(self):
        embedding_1 = get_embedding_size(10)
        self.assertEqual(embedding_1, 5)

        embedding_2 = get_embedding_size(15)
        self.assertEqual(embedding_2, 8)

        embedding_3 = get_embedding_size(20)
        self.assertEqual(embedding_3, 10)


@staticmethod
def make_default_dataframe():
    return pd.DataFrame(np.random.randint(0, 10, size=(10, 3)), columns=list('ABC'))


class CustomProcessor(TargetProcessor):
    def process_target(self, y: List) -> np.ndarray:
        # just for example purposes, let's use a LabelEncoder
        return LabelEncoder().fit_transform(y)


class CustomAssembler(ModelAssembler):
    def make_final_layer(self, previous_layer: Layer):
        output_model = Dense(1)(previous_layer)
        output_model = Activation('sigmoid')(output_model)
        return output_model

    def compile_model(self, model):
        model.compile(loss='binary_crossentropy', optimizer='adam', metrics=['accuracy'])
        return model

    """
    You can aso customize the hidden layers of the network
    """

    def make_hidden_layers(self, outputs: List[Layer]):
        output_model = Concatenate()(outputs)
        output_model = Dense(5000, kernel_initializer="uniform")(output_model)
        output_model = Activation('relu')(output_model)
        output_model = Dense(1000, kernel_initializer="uniform")(output_model)
        output_model = Activation('relu')(output_model)
        return output_model



import os

import pandas as pd

# from entity_embeddings.network.assembler import ModelAssembler
# from entity_embeddings.processor.processor import TargetProcessor
from entity_embeddings.network import ModelAssembler
from entity_embeddings.processor.processor import TargetProcessor


def check_csv_data(csv_path: str) -> None:
    if not csv_path:
        raise ValueError("You should provide a csv path where the data is located")

    if not os.path.exists(csv_path):
        raise ValueError("You should provide an existent csv path")


def check_not_empty_dataframe(df: pd.DataFrame) -> None:
    if df.empty:
        raise ValueError("You should provide a non-empty pandas dataframe")


def check_target_name(target_name: str) -> None:
    if not target_name:
        raise ValueError("You should provide a non-empty target name")


def check_target_existent_in_df(target_name: str, df: pd.DataFrame) -> None:
    if target_name not in df.columns:
        raise ValueError("You should provide a target variable that is existent on the dataframe")


def check_train_ratio(train_ratio: float) -> None:
    if train_ratio <= 0 or train_ratio >= 1:
        raise ValueError("You should provide a train ratio greater than 0 and smaller than 1")


def check_epochs(epochs: int) -> None:
    if epochs <= 0:
        raise ValueError("You should provide a epoch greater than zero")


def check_batch_size(batch_size: int) -> None:
    if batch_size <= 0:
        raise ValueError("You should provide a batch size greater than zero")


def check_weights_output(weights_output: str) -> None:
    if not weights_output:
        raise ValueError("You should provide a output file for the embeddings weights")

def check_target_processor(processor: TargetProcessor) -> None:
    if not isinstance(processor, TargetProcessor):
        raise ValueError("You should provide a target processor that inherits from TargetProcessor")

def check_model_assembler(assembler: ModelAssembler) -> None:
    if not isinstance(assembler, ModelAssembler):
        raise ValueError("You should provide a model assembler that inherits from ModelAssembler")


import unittest

import numpy as np

from entity_embeddings.util import dataframe_utils
from entity_embeddings.util.preprocessing_utils import transpose_to_list, series_to_list, get_X_y


class TestPreprocessingUtils(unittest.TestCase):
    def test_transpose_to_list(self):
        X_array = np.random.randint(0, 10, size=(10, 4))

        feature_list = transpose_to_list(X_array)

        self.assertEqual(len(feature_list), X_array.shape[1])
        self.__check_items(X_array, feature_list)

    def test_series_to_list(self):
        df = dataframe_utils.create_random_dataframe(rows=5, cols=5, columns='ABCDE')

        for row in df.iterrows():
            output_list = series_to_list(row)
            self.assertListEqual(row[1].values.tolist(), output_list[1].values.tolist())

    def test_get_X_y(self):
        df = dataframe_utils.create_random_dataframe(rows=5, cols=5, columns='ABCDE')
        target = 'E'

        X_list, y_list = get_X_y(df, target)

        # check y values
        for index, y_final in enumerate(y_list):
            self.assertEqual(df.loc[:, df.columns == target].values.tolist()[index][0], y_final)

        for index_list, X_list in enumerate(X_list):
            for index_item, item_x in enumerate(X_list):
                current_df = df.loc[:, df.columns != target].values.tolist()[index_list][index_item]

                self.assertEqual(current_df, item_x)

    def __check_items(self, X_array, feature_list):
        for index in range(X_array.shape[1]):
            for item_index, item_value in enumerate(X_array[index:index + 1, :-1].tolist()[0]):
                self.assertEqual(item_value, feature_list[item_index][index])



"""
This file contains both the Category and EmbeddingConfig classes, which are responsible to store data related to the
categories. This data will be later on used on our EmbeddingNetwork class.
"""
from typing import List

import numpy as np

from entity_embeddings.network.assembler import ModelAssembler, get_model_assembler
from entity_embeddings.processor.processor import TargetProcessor
from entity_embeddings.processor.target_type import TargetType
from entity_embeddings.util.dataframe_utils import load_guarantee_not_empty
from entity_embeddings.util.processor_utils import get_target_processor
from entity_embeddings.util.validation_utils import *


def get_embedding_size(unique_values: int) -> int:
    """
    This method is used to generate the embedding size to be used on our Embedding layer
    :param unique_values: the number of unique values in the given category
    :return:
    """
    return int(min(np.ceil(unique_values / 2), 50))


def generate_categories_from_df(df: pd.DataFrame, target_name: str):
    category_list = []

    for category in df:
        if not category == target_name:
            category_list.append(Category(category, df[category].nunique()))

    return category_list


class Category:
    """
    This class is used to store data related to the used category, such as the embedding size to be used
    """

    def __init__(self, alias: str, unique_values: int):
        self.alias = alias
        self.unique_values = unique_values
        self.embedding_size = get_embedding_size(unique_values)


class Config:
    """
    This class is used to store all the configuration (dataframes, target type, epochs, artifacts..) which will be
    used on our Embeddings Network
    """

    def __init__(self,
                 csv_path: str,
                 target_name: str,
                 train_ratio: float,
                 target_processor: TargetProcessor,
                 model_assembler: ModelAssembler,
                 epochs: int = 10,
                 batch_size: int = 128,
                 verbose: bool = False,
                 weights_output: str = 'weights_embeddings.pickle'):
        # input validations
        check_csv_data(csv_path)
        check_target_name(target_name)
        check_train_ratio(train_ratio)
        check_epochs(epochs)
        check_batch_size(batch_size)
        check_weights_output(weights_output)

        check_target_processor(target_processor)
        check_model_assembler(model_assembler)
        # TODO check processor and assembler

        self.csv_path = csv_path
        self.target_name = target_name
        self.train_ratio = train_ratio
        self.epochs = epochs
        self.batch_size = batch_size
        self.verbose = verbose
        self.weights_output = weights_output
        self.target_processor = target_processor
        self.model_assembler = model_assembler

        self.df = load_guarantee_not_empty(self.csv_path)
        check_target_existent_in_df(self.target_name, self.df)

        self.unique_classes = self.df[self.target_name].nunique()

        self.categories: List[Category] = generate_categories_from_df(self.df, self.target_name)

    @classmethod
    def make_default_config(cls,
                            csv_path: str,
                            target_name: str,
                            target_type: TargetType,
                            train_ratio: float,
                            epochs: int = 10,
                            batch_size: int = 128,
                            verbose: bool = False,
                            weights_output: str = 'weights_embeddings.pickle'):
        df = load_guarantee_not_empty(csv_path)
        check_target_existent_in_df(target_name, df)
        n_unique_classes = df[target_name].nunique()

        target_processor = get_target_processor(target_type)
        model_assembler = get_model_assembler(target_type, n_unique_classes)

        return cls(csv_path,
                   target_name,
                   train_ratio,
                   target_processor,
                   model_assembler,
                   epochs,
                   batch_size,
                   verbose,
                   weights_output)

    @classmethod
    def make_custom_config(cls,
                           csv_path: str,
                           target_name: str,
                           train_ratio: float,
                           target_processor: TargetProcessor,
                           model_assembler: ModelAssembler,
                           epochs: int = 10,
                           batch_size: int = 128,
                           verbose: bool = False,
                           weights_output: str = 'weights_embeddings.pickle'):
        return cls(csv_path,
                   target_name,
                   train_ratio,
                   target_processor,
                   model_assembler,
                   epochs,
                   batch_size,
                   verbose,
                   weights_output)



import unittest

import keras
from keras import Input, Model
from keras.engine import Layer
from keras.optimizers import Optimizer

from entity_embeddings.config import Config
from entity_embeddings.processor.target_type import TargetType
from entity_embeddings.util.dataframe_utils import create_random_csv, remove_random_csv


class TestAssembler(unittest.TestCase):
    def tearDown(self):
        remove_random_csv()

    def make_config_for_type(self, type: TargetType):
        csv_path = create_random_csv(5, 5, 'ABCDE')
        config = Config.make_default_config(csv_path=csv_path,
                                            target_name='E',
                                            target_type=type,
                                            train_ratio=0.9)

        return config

    def make_sample_layer(self):
        input_layer = Input(shape=(1,))
        return input_layer

    def test_activation_for_binary_is_sigmoid(self):
        config = self.make_config_for_type(TargetType.BINARY_CLASSIFICATION)
        previous_layer = self.make_sample_layer()

        layer = config.model_assembler.make_final_layer(previous_layer)

        self.check_layer_properties(layer=layer,
                                    activation="Sigmoid",
                                    outputs=1,
                                    operation='dense')

    def test_activation_for_multiclass_is_softmax(self):
        config = self.make_config_for_type(TargetType.MULTICLASS_CLASSIFICATION)
        previous_layer = self.make_sample_layer()

        layer = config.model_assembler.make_final_layer(previous_layer)
        self.check_layer_properties(layer=layer,
                                    activation="Softmax",
                                    outputs=config.unique_classes,
                                    operation='dense')

    def test_activation_for_regression_is_identity(self):
        config = self.make_config_for_type(TargetType.REGRESSION)
        previous_layer = self.make_sample_layer()

        layer = config.model_assembler.make_final_layer(previous_layer)
        self.check_layer_properties(layer=layer,
                                    activation="Identity",
                                    outputs=1,
                                    operation='dense')

    def test_model_params_for_binary_classification(self):
        config = self.make_config_for_type(TargetType.BINARY_CLASSIFICATION)
        previous_layer = self.make_sample_layer()

        layer = config.model_assembler.make_final_layer(previous_layer)
        model = Model(inputs=previous_layer, outputs=layer)

        model = config.model_assembler.compile_model(model)
        self.check_model_parameters(model=model,
                                    optimizer=keras.optimizers.Adam,
                                    loss="binary_crossentropy",
                                    metrics=["accuracy"])

    def test_model_params_for_multiclass_classification(self):
        config = self.make_config_for_type(TargetType.MULTICLASS_CLASSIFICATION)
        previous_layer = self.make_sample_layer()

        layer = config.model_assembler.make_final_layer(previous_layer)
        model = Model(inputs=previous_layer, outputs=layer)

        model = config.model_assembler.compile_model(model)
        self.check_model_parameters(model=model,
                                    optimizer=keras.optimizers.Adam,
                                    loss="categorical_crossentropy",
                                    metrics=[])

    def test_model_params_for_regression_classification(self):
        config = self.make_config_for_type(TargetType.REGRESSION)
        previous_layer = self.make_sample_layer()

        layer = config.model_assembler.make_final_layer(previous_layer)
        model = Model(inputs=previous_layer, outputs=layer)

        model = config.model_assembler.compile_model(model)
        self.check_model_parameters(model=model,
                                    optimizer=keras.optimizers.Adam,
                                    loss="mse",
                                    metrics=[])

    def check_model_parameters(self, model: Model, optimizer: Optimizer, loss: str, metrics) -> None:
        self.assertIsInstance(model.optimizer, optimizer)
        self.assertEqual(model.loss, loss)
        self.assertEqual(model.metrics, metrics)

    def check_layer_properties(self, layer: Layer, activation: str, outputs: int, operation: str):
        self.assertEqual(layer.op.type, activation)

        # TODO move those things to layer_utils class
        n_output = layer.shape[1].value
        self.assertEqual(n_output, outputs)

        op = layer.op._inputs[0].op.name.split("_")[0]
        self.assertEqual(op, operation)



from abc import ABC, abstractmethod
from typing import List

from keras.engine import Layer
from keras.layers import Dense, Activation, Concatenate
from keras.models import Model as KerasModel

from entity_embeddings.processor.target_type import TargetType


def get_model_assembler(target_type: TargetType, n_unique_classes: int):
    if target_type == TargetType.BINARY_CLASSIFICATION:
        return BinaryClassificationAssembler()
    elif target_type == TargetType.MULTICLASS_CLASSIFICATION:
        return MulticlassClassificationAssembler(n_unique_classes)
    elif target_type == TargetType.REGRESSION:
        return RegressionClassificationAssembler()


class ModelAssembler(ABC):
    @abstractmethod
    def make_final_layer(self, previous_layer: Layer) -> Layer:
        raise NotImplementedError("Your model assembler should override the method make_final_layer")

    @abstractmethod
    def compile_model(self, model: KerasModel) -> KerasModel:
        raise NotImplementedError("Your model assembler should override the method compile_model")

    def make_hidden_layers(self, outputs: List[Layer]) -> Layer:
        output_model = Concatenate()(outputs)
        output_model = Dense(1000, kernel_initializer="uniform")(output_model)
        output_model = Activation('relu')(output_model)
        output_model = Dense(500, kernel_initializer="uniform")(output_model)
        output_model = Activation('relu')(output_model)
        return output_model


class BinaryClassificationAssembler(ModelAssembler):
    def make_final_layer(self, previous_layer: Layer) -> Layer:
        output_model = Dense(1)(previous_layer)
        output_model = Activation('sigmoid')(output_model)
        return output_model

    def compile_model(self, model):
        model.compile(loss='binary_crossentropy', optimizer='adam', metrics=['accuracy'])
        return model


class MulticlassClassificationAssembler(ModelAssembler):
    def __init__(self, n_unique_classes: int):
        self.n_unique_classes = n_unique_classes

    def make_final_layer(self, previous_layer: Layer) -> Layer:
        output_model = Dense(self.n_unique_classes)(previous_layer)
        output_model = Activation('softmax')(output_model)
        return output_model

    def compile_model(self, model):
        model.compile(loss='categorical_crossentropy', optimizer='adam')
        return model


class RegressionClassificationAssembler(ModelAssembler):
    def make_final_layer(self, previous_layer: Layer) -> Layer:
        output_model = Dense(1)(previous_layer)
        output_model = Activation('linear')(output_model)
        return output_model

    def compile_model(self, model):
        model.compile(loss='mse', optimizer='adam')
        return model



import unittest
from unittest import mock
from unittest.mock import Mock

import numpy as np

from entity_embeddings.config import Config
from entity_embeddings.embedder import Embedder
from entity_embeddings.processor.target_type import TargetType
from entity_embeddings.util.dataframe_utils import create_random_csv, remove_random_csv
from entity_embeddings.util.model_utils import get_weights_from_layer


class TestEmbedder(unittest.TestCase):
    def setUp(self):
        config = self.make_default_config()
        self.embedder = Embedder(config)

        self.embedder.prepare_data = Mock()

        self.embedder.prepare_data.return_value = [np.random.randint(0, 1, size=(10, 3)),
                                                   np.random.randint(0, 1, size=(10, 3)),
                                                   np.random.randint(0, 1, size=(10, 1)),
                                                   np.random.randint(0, 1, size=(10, 1))]

    def tearDown(self):
        remove_random_csv()

    def test_call_prepare_data_on_perform_embedding(self):
        self.embedder.perform_embedding()
        self.assertTrue(self.embedder.prepare_data.called)

    def make_default_config(self) -> Config:
        random_csv = create_random_csv()
        target = 'D'

        config = Config.make_default_config(csv_path=random_csv,
                                            target_name=target,
                                            target_type=TargetType.BINARY_CLASSIFICATION,
                                            train_ratio=0.9)

        return config



import unittest

import pandas as pd

from entity_embeddings.util import dataframe_utils
from entity_embeddings.util.validation_utils import check_csv_data, check_not_empty_dataframe, check_target_name, \
    check_target_existent_in_df, check_train_ratio, check_epochs, check_batch_size, check_weights_output, \
    check_target_processor, check_model_assembler


class TestValidationUtils(unittest.TestCase):
    def test_check_csv_data_not_existent(self) -> None:
        csv_path = 'not_existent_path.csv'
        self.assertRaises(ValueError, check_csv_data, csv_path)

    def test_check_csv_data_empty(self) -> None:
        csv_path = ''
        self.assertRaises(ValueError, check_csv_data, csv_path)

    def test_check_not_empty_dataframe(self) -> None:
        df = pd.DataFrame()
        self.assertRaises(ValueError, check_not_empty_dataframe, df)

    def test_check_target_name(self) -> None:
        target_name = ''
        self.assertRaises(ValueError, check_target_name, target_name)

    def test_check_target_existent_in_df(self) -> None:
        df = dataframe_utils.create_random_dataframe(columns='ABCD')
        self.assertRaises(ValueError, check_target_existent_in_df, 'E', df)

    def test_check_train_ratio_greater_than_one(self) -> None:
        ratio = 1.1
        self.assertRaises(ValueError, check_train_ratio, ratio)

    def test_check_train_ratio_equal_one(self) -> None:
        ratio = 1
        self.assertRaises(ValueError, check_train_ratio, ratio)

    def test_check_train_ratio_smaller_than_zero(self) -> None:
        ratio = -0.1
        self.assertRaises(ValueError, check_train_ratio, ratio)

    def test_check_train_ratio_equals_zero(self) -> None:
        ratio = 0.0
        self.assertRaises(ValueError, check_train_ratio, ratio)

    def test_check_epochs(self) -> None:
        epochs = 0
        self.assertRaises(ValueError, check_epochs, epochs)

    def test_check_batch_size(self) -> None:
        batch_size = 0
        self.assertRaises(ValueError, check_batch_size, batch_size)

    def test_check_weights_output(self) -> None:
        weights_output = ''
        self.assertRaises(ValueError, check_weights_output, weights_output)

    def test_check_target_processor(self) -> None:
        invalid_processor = ()
        self.assertRaises(ValueError, check_target_processor, invalid_processor)

    def test_check_model_assembler(self) -> None:
        invalid_assembler = ()
        self.assertRaises(ValueError, check_model_assembler, invalid_assembler)



import pickle

from keras import Model
from keras.engine import Layer
from keras.layers import Dense, Activation

from entity_embeddings import config
from entity_embeddings.processor.target_type import TargetType
from entity_embeddings.util import dataframe_utils


def save_weights(model: Model, config: config) -> None:
    weights_embeddings = []
    for column in dataframe_utils.get_all_columns_except(config.df, config.target_name):
        weights = get_weights_from_layer(model, column)
        weights_embeddings.append(weights)

    with open(config.weights_output, 'wb') as f:
        pickle.dump(weights_embeddings, f, -1)


def get_weights_from_layer(model, layer_name):
    return model.get_layer(layer_name).get_weights()[0]



import os
import shutil

import numpy as np
import pandas as pd

from entity_embeddings.util.validation_utils import check_not_empty_dataframe


def load_guarantee_not_empty(csv_path) -> pd.DataFrame:
    df = pd.read_csv(csv_path)
    check_not_empty_dataframe(df)

    return df


def get_all_columns_except(df: pd.DataFrame, column_to_skip: str) -> pd.DataFrame:
    return df.loc[:, df.columns != column_to_skip]


# Below are some methods that are used mostly for testing purposes

BASE_OUTPUT_DIR = 'temp_artifacts'
BASE_OUTPUT_FILENAME = 'temp_dataframe.csv'


def create_random_dataframe(rows: int = 4,
                            cols: int = 4,
                            columns: str = 'ABCD') -> pd.DataFrame:
    return pd.DataFrame(np.random.randint(0, 10, size=(rows, cols)), columns=list(columns))


def create_random_csv(rows: int = 4,
                      cols: int = 4,
                      columns: str = 'ABCD') -> str:
    df = create_random_dataframe(rows, cols, columns)

    full_path = os.path.join(BASE_OUTPUT_DIR, BASE_OUTPUT_FILENAME)
    os.makedirs(BASE_OUTPUT_DIR, exist_ok=True)

    df.to_csv(full_path, index=False)

    return full_path


def remove_random_csv() -> None:
    shutil.rmtree(BASE_OUTPUT_DIR)



import os
import unittest

import pandas as pd

from entity_embeddings.util.dataframe_utils import create_random_dataframe, create_random_csv, remove_random_csv, \
    get_all_columns_except
from entity_embeddings.util.validation_utils import check_not_empty_dataframe


class TestDataframeUtils(unittest.TestCase):
    def test_create_random_dataframe(self):
        rows = 5
        cols = 5
        columns = 'ABCDE'

        df = create_random_dataframe(rows, cols, columns)

        self.__check_dataframe_data(df, rows, cols, columns)

    def test_create_and_remove_random_csv(self):
        rows = 5
        cols = 5
        columns = 'ABCDE'

        csv_path = create_random_csv(rows, cols, columns)

        exists_csv = os.path.exists(csv_path)
        self.assertEqual(exists_csv, True)

        df = pd.read_csv(csv_path)
        self.__check_dataframe_data(df, rows, cols, columns)

        remove_random_csv()

        exists_csv = os.path.exists(csv_path)
        self.assertEqual(exists_csv, False)

    def test_load_guarantee_not_empty(self):
        df = pd.DataFrame()
        self.assertRaises(ValueError, check_not_empty_dataframe, df)

    def test_get_all_columns_except(self):
        columns = 'ABCD'
        to_remove = 'D'
        df = create_random_dataframe(rows=4, cols=4, columns=columns)

        df = get_all_columns_except(df, to_remove)
        self.assertEqual(''.join(list(df)), columns.replace(to_remove, ''))

    def __check_dataframe_data(self, df: pd.DataFrame, rows: int, cols: int, columns: str) -> None:
        self.assertEqual(df.shape[0], rows)
        self.assertEqual(df.shape[1], cols)
        self.assertEqual(''.join(list(df)), columns)



"""
This file contains the implementation of the class EmbeddingNetwork, which will be responsible for the initialization
and setup of our entity embedding network
"""
from typing import Tuple, List

import numpy as np
from keras import Input
from keras.engine import Layer
from keras.layers import Concatenate, Embedding, Reshape
from keras.layers import Dense, Activation
from keras.models import Model as KerasModel

from entity_embeddings.config import Config
from entity_embeddings.util.preprocessing_utils import transpose_to_list

np.random.seed(42)


class EmbeddingNetwork:
    """
    This class is used to provide a Entity Embedding Network from a given EmbeddingConfig object
    """

    def __init__(self, config: Config):
        super().__init__()
        self.config = config
        self.model = self.__make_model()

    def __make_model(self) -> KerasModel:
        """
        This method is used to generate our KerasModel containing the Embedding layers alongside with the output layers
        :return: a compiled KerasModel object
        """

        inputs, outputs = self._make_embedding_layers()
        output_model = self.config.model_assembler.make_hidden_layers(outputs)
        output_model = self.config.model_assembler.make_final_layer(output_model)

        model = KerasModel(inputs=inputs, outputs=output_model)
        model = self.config.model_assembler.compile_model(model)
        return model


    def _make_embedding_layers(self) -> Tuple[List[Layer], List[Layer]]:
        """
        This method is used to generate the list of inputs and output layers where our Embedding layers will be placed
        :return: a tuple containing two lists: the first, the input layers; the second, the output layers
        """
        embedding_inputs = []
        embedding_outputs = []

        for category in self.config.categories:
            input_category = Input(shape=(1,))
            output_category = Embedding(input_dim=category.unique_values,
                                        output_dim=category.embedding_size,
                                        name=category.alias)(input_category)
            output_category = Reshape(target_shape=(category.embedding_size,))(output_category)

            embedding_inputs.append(input_category)
            embedding_outputs.append(output_category)

        return embedding_inputs, embedding_outputs

    def fit(self, X_train: np.ndarray, y_train: np.ndarray, X_val: np.ndarray, y_val: np.ndarray) -> None:
        """
        This method is used to fit a given training and validation data into our entity embeddings model
        :param X_train: training features
        :param y_train: training targets
        :param X_val: validation features
        :param y_val: validation targets
        """
        self.model.fit(x=transpose_to_list(X_train),
                       y=y_train,
                       validation_data=(transpose_to_list(X_val), y_val),
                       epochs=self.config.epochs,
                       batch_size=self.config.batch_size)



from typing import List, Tuple

import numpy as np
import pandas as pd
from sklearn import preprocessing


def series_to_list(series: pd.Series) -> List:
    """
    This method is used to convert a given pd.Series object into a list
    :param series: the list to be converted
    :return: the list containing all the elements from the Series object
    """
    list_cols = []
    for item in series:
        list_cols.append(item)

    return list_cols


def sample(X: np.ndarray, y: np.ndarray, n: int) -> Tuple[np.ndarray, np.ndarray]:
    num_row = X.shape[0]
    indices = np.random.randint(num_row, size=n)
    return X[indices, :], y[indices]


def get_X_y(df: pd.DataFrame, name_target: str) -> Tuple[List, List]:
    """
    This method is used to gather the X (features) and y (targets) from a given dataframe based on a given
    target name
    :param df: the dataframe to be used as source
    :param name_target: the name of the target variable
    :return: the list of features and targets
    """
    X_list = []
    y_list = []

    for index, record in df.iterrows():
        fl = series_to_list(record.drop(name_target))
        X_list.append(fl)
        y_list.append(int(record[name_target]))

    return X_list, y_list


def label_encode(data: List) -> np.ndarray:
    """
    This method is used to perform Label Encoding on a given list
    :param data: the list containing the items to be encoded
    :return: the encoded np.ndarray
    """
    data_encoded = np.array(data)
    for i in range(data_encoded.shape[1]):
        le = preprocessing.LabelEncoder()
        le.fit(data_encoded[:, i])
        data_encoded[:, i] = le.transform(data_encoded[:, i])
    data_encoded = data_encoded.astype(int)
    return data_encoded


def transpose_to_list(X: np.ndarray) -> List[np.ndarray]:
    """
    :param X: the ndarray to be used as source
    :return: a list of nd.array containing the elements from the numpy array
    """

    features_list = []
    for index in range(X.shape[1]):
        features_list.append(X[..., [index]])

    return features_list



import unittest

from keras.layers import Embedding

from entity_embeddings.config import Config, get_embedding_size
from entity_embeddings.network.network import EmbeddingNetwork
from entity_embeddings.processor.target_type import TargetType
from entity_embeddings.util.dataframe_utils import create_random_csv, remove_random_csv


class TestNetwork(unittest.TestCase):
    def test_model_embedding_size(self):

        random_csv = create_random_csv()
        target = 'D'

        config = Config.make_default_config(csv_path=random_csv,
                                            target_name=target,
                                            target_type=TargetType.BINARY_CLASSIFICATION,
                                            train_ratio=0.9)

        network = EmbeddingNetwork(config)

        for layer in network.model.layers:
            if type(layer) is Embedding:
                embedding_size = int(layer.embeddings.initial_value.shape[1])
                self.assertEqual(get_embedding_size(config.df[layer.name].nunique()), embedding_size)

        remove_random_csv()

    def test_output_for_regression(self):
        pass

    def test_output_for_binary_classification(self):
        pass
